/*
 * dump-stools.c --
 *
 *      Operations to generate MIB module stubs for the stools package
 *
 * Copyright (c) 2001 J. Schoenwaelder, Technical University of Braunschweig.
 *
 * See the file "COPYING" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * @(#) $Id: dump-stools.c,v 1.11 2001/03/05 14:27:42 schoenw Exp $
 */

/*
 * TODO:
 *	  - generate #defines for deprecated and obsolete objects
 *	  - generate stub codes for the various functions
 *	  - generate type and range checking code
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_WIN_H
#include "win.h"
#endif

#include "smi.h"
#include "smidump.h"



static char *getStringTime(time_t t)
{
    static char   s[27];
    struct tm	  *tm;

    tm = gmtime(&t);
    sprintf(s, "%04d-%02d-%02d %02d:%02d",
	    tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
	    tm->tm_hour, tm->tm_min);
    return s;
}



static void
printTopComment(FILE *f, SmiModule *smiModule)
{
    SmiRevision *smiRevision;
    char *revision = "[unknown]";

    smiRevision = smiGetFirstRevision(smiModule);
    if (smiRevision) {
	revision = getStringTime(smiRevision->date);
    }

    fprintf(f,
	    "/*	\t\t\t\t\t\t-- DO NOT EDIT --\n"
	    " * This file has been generated by smidump\n"
	    " * version " SMI_VERSION_STRING " for the stools package.\n"
	    " *\n"
	    " * Derived from %s revision %s.\n"
	    " *\n * $I" "d$\n"
	    " */\n"
	    "\n", smiModule->name, revision);
}



static char* translate(char *m)
{
    char *s;
    int i;

    s = xstrdup(m);
    for (i = 0; s[i]; i++) {
	if (s[i] == '-') s[i] = '_';
    }
  
    return s;
}



static char* translateUpper(char *m)
{
    char *s;
    int i;

    s = xstrdup(m);
    for (i = 0; s[i]; i++) {
	if (s[i] == '-') s[i] = '_';
	if (islower((int) s[i])) {
	    s[i] = toupper(s[i]);
	}
    }
  
    return s;
}



static char* translateLower(char *m)
{
    char *s;
    int i;

    s = xstrdup(m);
    for (i = 0; s[i]; i++) {
	if (s[i] == '-') s[i] = '_';
	if (isupper((int) s[i])) {
	    s[i] = tolower(s[i]);
	}
    }
  
    return s;
}



static char* translateFileName(char *m)
{
    char *s;
    int i;

    s = xstrdup(m);
    for (i = 0; s[i]; i++) {
	if (s[i] == '_') s[i] = '-';
	if (isupper((int) s[i])) {
	    s[i] = tolower(s[i]);
	}
    }
  
    return s;
}



static FILE * createFile(char *name, char *suffix)
{
    char *fullname;
    FILE *f;

    fullname = xmalloc(strlen(name) + (suffix ? strlen(suffix) : 0) + 2);
    strcpy(fullname, name);
    if (suffix) {
        strcat(fullname, suffix);
    }
    if (!access(fullname, R_OK)) {
        fprintf(stderr, "smidump: %s already exists\n", fullname);
        xfree(fullname);
        return NULL;
    }
    f = fopen(fullname, "w");
    if (!f) {
        fprintf(stderr, "smidump: cannot open %s for writing: ", fullname);
        perror(NULL);
        xfree(fullname);
        exit(1);
    }
    xfree(fullname);
    return f;
}



static int isGroup(SmiNode *smiNode)
{
    SmiNode *childNode;

    if (smiNode->nodekind == SMI_NODEKIND_ROW) {
	return 1;
    }
    
    for (childNode = smiGetFirstChildNode(smiNode);
	 childNode;
	 childNode = smiGetNextChildNode(childNode)) {
	if (childNode->nodekind == SMI_NODEKIND_SCALAR) {
	    return 1;
	}
    }

    return 0;
}



static int isAccessible(SmiNode *groupNode)
{
    SmiNode *smiNode;
    int num = 0;
    
    for (smiNode = smiGetFirstChildNode(groupNode);
	 smiNode;
	 smiNode = smiGetNextChildNode(smiNode)) {
	if ((smiNode->nodekind == SMI_NODEKIND_SCALAR
	     || smiNode->nodekind == SMI_NODEKIND_COLUMN)
	    && (smiNode->access == SMI_ACCESS_READ_ONLY
		|| smiNode->access == SMI_ACCESS_READ_WRITE)) {
	    num++;
	}
    }

    return num;
}



static int isIndex(SmiNode *groupNode, SmiNode *smiNode)
{
    SmiElement *smiElement;
    
    /*
     * Perhaps this test needs to be more sophisticated if you have
     * really creative cross-table indexing constructions...
     */

    for (smiElement = smiGetFirstElement(groupNode);
	 smiElement; smiElement = smiGetNextElement(smiElement)) {
	if (smiNode == smiGetElementNode(smiElement)) {
	    return 1;
	}
    }

    return 0;
}



static unsigned int getMinSize(SmiType *smiType)
{
    SmiRange *smiRange;
    SmiType  *parentType;
    unsigned int min = 65535, size;
    
    switch (smiType->basetype) {
    case SMI_BASETYPE_BITS:
	return 0;
    case SMI_BASETYPE_OCTETSTRING:
    case SMI_BASETYPE_OBJECTIDENTIFIER:
	size = 0;
	break;
    default:
	return -1;
    }

    for (smiRange = smiGetFirstRange(smiType);
	 smiRange ; smiRange = smiGetNextRange(smiRange)) {
	if (smiRange->minValue.value.unsigned32 < min) {
	    min = smiRange->minValue.value.unsigned32;
	}
    }
    if (min < 65535 && min > size) {
	size = min;
    }

    parentType = smiGetParentType(smiType);
    if (parentType) {
	unsigned int psize = getMinSize(parentType);
	if (psize > size) {
	    size = psize;
	}
    }

    return size;
}



static unsigned int getMaxSize(SmiType *smiType)
{
    SmiRange *smiRange;
    SmiType  *parentType;
    SmiNamedNumber *nn;
    unsigned int max = 0, size;
    
    switch (smiType->basetype) {
    case SMI_BASETYPE_BITS:
    case SMI_BASETYPE_OCTETSTRING:
	size = 65535;
	break;
    case SMI_BASETYPE_OBJECTIDENTIFIER:
	size = 128;
	break;
    default:
	return -1;
    }

    if (smiType->basetype == SMI_BASETYPE_BITS) {
	for (nn = smiGetFirstNamedNumber(smiType);
	     nn;
	     nn = smiGetNextNamedNumber(nn)) {
	    if (nn->value.value.unsigned32 > max) {
		max = nn->value.value.unsigned32;
	    }
	}
	size = (max / 8) + 1;
	return size;
    }

    for (smiRange = smiGetFirstRange(smiType);
	 smiRange ; smiRange = smiGetNextRange(smiRange)) {
	if (smiRange->maxValue.value.unsigned32 > max) {
	    max = smiRange->maxValue.value.unsigned32;
	}
    }
    if (max > 0 && max < size) {
	size = max;
    }

    parentType = smiGetParentType(smiType);
    if (parentType) {
	unsigned int psize = getMaxSize(parentType);
	if (psize < size) {
	    size = psize;
	}
    }

    return size;
}


static void printHeaderEnums(FILE *f, SmiModule *smiModule)
{
    SmiNode  *smiNode;
    SmiType  *smiType;
    char     *cName, *cModuleName;
    int      cnt = 0;

    cModuleName = translateLower(smiModule->name);

    for (smiNode = smiGetFirstNode(smiModule,
				   SMI_NODEKIND_SCALAR | SMI_NODEKIND_COLUMN);
	 smiNode;
	 smiNode = smiGetNextNode(smiNode,
				  SMI_NODEKIND_SCALAR | SMI_NODEKIND_COLUMN)) {
	smiType = smiGetNodeType(smiNode);
	if (smiType && smiType->basetype == SMI_BASETYPE_ENUM) {
	    if (cnt == 0) {
		fprintf(f,
			"/*\n"
			" * Tables to map enumerations to strings and vice versa.\n"
			" */\n"
			"\n");
	    }
	    cnt++;
	    cName = translate(smiNode->name);
	    fprintf(f, "extern stls_table_t %s_enums_%s[];\n",
		    cModuleName, cName);
	    xfree(cName);
	}
    }
    
    if (cnt) {
	fprintf(f, "\n");
    }

    xfree(cModuleName);
}



static void printHeaderTypedefMember(FILE *f, SmiNode *smiNode,
				     SmiType *smiType, int isIndex)
{
    char *cName;
    unsigned minSize, maxSize;

    cName = translate(smiNode->name);
    switch (smiType->basetype) {
    case SMI_BASETYPE_OBJECTIDENTIFIER:
	maxSize = getMaxSize(smiType);
	minSize = getMinSize(smiType);
	if (isIndex) {
	    fprintf(f,
		    "    guint32  %s[%u];\n", cName,
		    maxSize < 128 ? maxSize : 128);
	} else {
	    fprintf(f,
		    "    guint32  *%s;\n", cName);
	}
	if (maxSize != minSize) {
	    fprintf(f,
		    "    gsize    _%sLength;\n", cName);
	}
	break;
    case SMI_BASETYPE_OCTETSTRING:
    case SMI_BASETYPE_BITS:
	maxSize = getMaxSize(smiType);
	minSize = getMinSize(smiType);
	if (isIndex) {
	    fprintf(f,
		    "    guchar   %s[%u];\n", cName,
		    maxSize < 128 ? maxSize : 128);
	} else {
	    fprintf(f,
		    "    guchar   *%s;\n", cName);
	}
	if (maxSize != minSize) {
	    fprintf(f,
		    "    gsize    _%sLength;\n", cName);
	}
	break;
    case SMI_BASETYPE_ENUM:
    case SMI_BASETYPE_INTEGER32:
	fprintf(f,
		"    gint32   %s%s;\n", isIndex ? "" : "*", cName);
	break;
    case SMI_BASETYPE_UNSIGNED32:
	fprintf(f,
		"    guint32  %s%s;\n", isIndex ? "" : "*", cName);
	break;
    case SMI_BASETYPE_INTEGER64:
	fprintf(f,
		"    gint64   *%s; \n", cName);
	break;
    case SMI_BASETYPE_UNSIGNED64:
	fprintf(f,
		"    guint64  *%s; \n", cName);
	break;
    default:
	fprintf(f,
		"    /* ?? */  _%s; \n", cName);
	break;
    }
    xfree(cName);
}



static void printHeaderTypedefIndex(FILE *f, SmiNode *smiNode, SmiNode *groupNode)
{
    SmiElement *smiElement;
    SmiNode *iNode;
    SmiType *iType;
    
    for (smiElement = smiGetFirstElement(smiNode);
	 smiElement; smiElement = smiGetNextElement(smiElement)) {
	iNode = smiGetElementNode(smiElement);
	if (iNode) {
	    iType = smiGetNodeType(iNode);
	    if (iType) {
		printHeaderTypedefMember(f, iNode, iType, 1);
	    }
	}
    }
}



static void printHeaderTypedef(FILE *f, SmiModule *smiModule,
			       SmiNode *groupNode)
{
    SmiNode *smiNode, *indexNode;
    SmiType *smiType;
    char    *cModuleName, *cGroupName;
    int     writable = 0;

    cModuleName = translateLower(smiModule->name);
    cGroupName = translate(groupNode->name);

    fprintf(f,
	    "/*\n"
	    " * C type definitions for %s::%s.\n"
	    " */\n\n",
	    smiModule->name, groupNode->name);
    
    fprintf(f, "typedef struct %s {\n", cGroupName);

    /*
     * print index objects that are not part of the group
     */

    switch (groupNode->indexkind) {
    case SMI_INDEX_INDEX:
    case SMI_INDEX_REORDER:
	printHeaderTypedefIndex(f, groupNode, groupNode);
	break;
    case SMI_INDEX_EXPAND:	/* TODO: we have to do more work here! */
	break;
    case SMI_INDEX_AUGMENT:
    case SMI_INDEX_SPARSE:
	indexNode = smiGetRelatedNode(groupNode);
	if (indexNode) {
	    printHeaderTypedefIndex(f, indexNode, groupNode);
	}
	break;
    case SMI_INDEX_UNKNOWN:
	break;
    }

	    
    for (smiNode = smiGetFirstChildNode(groupNode);
	 smiNode;
	 smiNode = smiGetNextChildNode(smiNode)) {
	if (smiNode->nodekind & (SMI_NODEKIND_COLUMN | SMI_NODEKIND_SCALAR)
	    && (smiNode->access >= SMI_ACCESS_READ_ONLY)) {
	    if (isIndex(groupNode, smiNode)) {
		continue;
	    }
	    if (smiNode->access == SMI_ACCESS_READ_WRITE) {
		writable++;
	    }
	    smiType = smiGetNodeType(smiNode);
	    if (! smiType) {
		continue;
	    }
	    printHeaderTypedefMember(f, smiNode, smiType, 0);
	}	    
    }

    fprintf(f, "} %s_t;\n\n", cGroupName);

    if (groupNode->nodekind == SMI_NODEKIND_ROW) {
	char *cTableName;
	SmiNode *tableNode;

	tableNode = smiGetParentNode(groupNode);
	if (tableNode) {
	    cTableName = translate(tableNode->name);
	    fprintf(f, "extern int\n"
		    "%s_get_%s(host_snmp *s, %s_t ***%s);\n\n",
		    cModuleName, cTableName, cGroupName, cGroupName);
	    fprintf(f, "extern void\n"
		    "%s_free_%s(%s_t **%s);\n\n",
		    cModuleName, cTableName, cGroupName, cGroupName);
	    xfree(cTableName);
	}
    }
    fprintf(f, "extern %s_t *\n"
	    "%s_new_%s();\n\n",
	    cGroupName, cModuleName, cGroupName);
    fprintf(f, "extern int\n"
	    "%s_get_%s(host_snmp *s, %s_t **%s);\n\n",
	    cModuleName, cGroupName, cGroupName, cGroupName);
    if (writable) {
	fprintf(f, "extern int\n"
		"%s_set_%s(host_snmp *s, %s_t *%s);\n\n",
		cModuleName, cGroupName, cGroupName, cGroupName);
    }
    fprintf(f, "extern void\n"
	    "%s_free_%s(%s_t *%s);\n\n",
	    cModuleName, cGroupName, cGroupName, cGroupName);
	    
    xfree(cGroupName);
    xfree(cModuleName);
}



static void printHeaderTypedefs(FILE *f, SmiModule *smiModule)
{
    SmiNode   *smiNode;
    int       cnt = 0;
    
    for (smiNode = smiGetFirstNode(smiModule, SMI_NODEKIND_ANY);
	 smiNode;
	 smiNode = smiGetNextNode(smiNode, SMI_NODEKIND_ANY)) {
	if (isGroup(smiNode) && isAccessible(smiNode)) {
	    cnt++;
	    printHeaderTypedef(f, smiModule, smiNode);
	}
    }
    
    if (cnt) {
	fprintf(f, "\n");
    }
}



static void dumpHeader(SmiModule *smiModule, char *baseName)
{
    char *pModuleName;
    FILE *f;
    
    pModuleName = translateUpper(smiModule->name);

    f = createFile(baseName, ".h");
    if (! f) {
	return;
    }

    printTopComment(f, smiModule);
    
    fprintf(f,
	    "#ifndef _%s_H_\n"
	    "#define _%s_H_\n"
	    "\n"
	    "#include \"stools.h\"\n"
	    "\n",
	    pModuleName, pModuleName);

    printHeaderEnums(f, smiModule);
    printHeaderTypedefs(f, smiModule);

    fprintf(f,
	    "#endif /* _%s_H_ */\n",
	    pModuleName);

    fclose(f);
    xfree(pModuleName);
}



static void printEnumTables(FILE *f, SmiModule *smiModule)
{
    SmiNode  *smiNode;
    SmiType  *smiType;
    SmiNamedNumber *nn;
    char     *cName, *cModuleName;
    int      cnt = 0;
    
    cModuleName = translateLower(smiModule->name);
    
    for (smiNode = smiGetFirstNode(smiModule,
				   SMI_NODEKIND_SCALAR | SMI_NODEKIND_COLUMN);
	 smiNode;
	 smiNode = smiGetNextNode(smiNode,
				  SMI_NODEKIND_SCALAR | SMI_NODEKIND_COLUMN)) {
	smiType = smiGetNodeType(smiNode);
	if (smiType && smiType->basetype == SMI_BASETYPE_ENUM) {
	    cnt++;
	    cName = translate(smiNode->name);
	    fprintf(f, "stls_table_t %s_enums_%s[] = {\n", cModuleName, cName);
	    for (nn = smiGetFirstNamedNumber(smiType); nn;
		 nn = smiGetNextNamedNumber(nn)) {
		fprintf(f, "    { %d, \"%s\" },\n",
			(int) nn->value.value.integer32, nn->name);
	    }
	    
	    fprintf(f,
		    "    { 0, NULL }\n"
		    "};\n"
		    "\n");
	    xfree(cName);
	}
    }
    
    if (cnt) {
	fprintf(f, "\n");
    }

    xfree(cModuleName);
    
}



static void printUnpackMethod(FILE *f, SmiNode *groupNode)
{
    SmiElement *smiElement;
    SmiNode *indexNode = NULL;
    SmiNode *iNode;
    SmiType *iType;
    char    *cGroupName, *cName;
    unsigned maxSize, minSize;
    int last = 0;
    
    cGroupName = translate(groupNode->name);

    switch (groupNode->indexkind) {
    case SMI_INDEX_INDEX:
    case SMI_INDEX_REORDER:
	indexNode = groupNode;
	break;
    case SMI_INDEX_EXPAND:	/* TODO: we have to do more work here! */
	indexNode = NULL;
	break;
    case SMI_INDEX_AUGMENT:
    case SMI_INDEX_SPARSE:
	indexNode = smiGetRelatedNode(groupNode);
	break;
    case SMI_INDEX_UNKNOWN:
	indexNode = NULL;
	break;
    }

    for (smiElement = smiGetFirstElement(indexNode);
	 smiElement; smiElement = smiGetNextElement(smiElement)) {
	iNode = smiGetElementNode(smiElement);
	if (iNode) {
	    iType = smiGetNodeType(iNode);
	    if (iType && iType->basetype == SMI_BASETYPE_OCTETSTRING) {
		break;
	    }
	}
    }

    fprintf(f,
	    "static int\n"
	    "unpack_%s(GSnmpVarBind *vb, %s_t *%s)\n"
	    "{\n"
	    "    int %sidx = %u;\n"
	    "\n",
	    cGroupName, cGroupName, cGroupName,
	    smiElement ? "i, len, " : "", groupNode->oidlen + 1);

    for (smiElement = smiGetFirstElement(indexNode);
	 smiElement; smiElement = smiGetNextElement(smiElement)) {
	iNode = smiGetElementNode(smiElement);
	last = (smiGetNextElement(smiElement) == NULL);
	if (iNode) {
	    iType = smiGetNodeType(iNode);
	    if (! iType) {
		continue;
	    }
	    cName = translate(iNode->name);
	    switch (iType->basetype) {
	    case SMI_BASETYPE_ENUM:
	    case SMI_BASETYPE_INTEGER32:
		fprintf(f,
			"    if (vb->id_len < idx) return -1;\n"
			"    %s->%s = vb->id[idx++];\n",
			cGroupName, cName);
		break;
	    case SMI_BASETYPE_UNSIGNED32:
		fprintf(f,
			"    if (vb->id_len < idx) return -1;\n"
			"    %s->%s = vb->id[idx++];\n",
			cGroupName, cName);
		break;
	    case SMI_BASETYPE_OCTETSTRING:
		maxSize = getMaxSize(iType);
		minSize = getMinSize(iType);
		if (minSize == maxSize) {
		    fprintf(f,
			    "    len = %u;\n"
			    "    if (vb->id_len < idx + len) return -1;\n"
			    "    for (i = 0; i < len; i++) {\n"
			    "        %s->%s[i] = vb->id[idx++];\n"
			    "    }\n",
			    minSize, cGroupName, cName);
		} else if (last && indexNode->implied) {
		    fprintf(f,
			    "    if (vb->id_len < idx) return -1;\n"
			    "    len = vb->id_len - idx;\n"
			    "    for (i = 0; i < len; i++) {\n"
			    "        %s->%s[i] = vb->id[idx++];\n"
			    "    }\n"
			    "    %s->_%sLength = len;\n",
			    cGroupName, cName, cGroupName, cName);
		} else {
		    fprintf(f,
			    "    if (vb->id_len < idx) return -1;\n"
			    "    len = vb->id[idx++];\n"
			    "    if (vb->id_len < idx + len) return -1;\n"
			    "    for (i = 0; i < len; i++) {\n"
			    "        %s->%s[i] = vb->id[idx++];\n"
			    "    }\n"
			    "    %s->_%sLength = len;\n",
			    cGroupName, cName, cGroupName, cName);
		}
		break;
	    default:
		fprintf(f,
			"    /* XXX how to unpack %s->%s ? */\n",
			cGroupName, cName);
		break;
	    }
	    xfree(cName);
	}
    }

    fprintf(f,
	    "    if (vb->id_len > idx) return -1;\n"
	    "    return 0;\n"
	    "}\n\n");

    xfree(cGroupName);
}



static void printVariableAssignement(FILE *f, SmiNode *groupNode)
{
    SmiNode *smiNode;
    SmiType *smiType;
    char    *cGroupName, *cName;
    unsigned maxSize, minSize;

    cGroupName = translate(groupNode->name);

    for (smiNode = smiGetFirstChildNode(groupNode);
	 smiNode;
	 smiNode = smiGetNextChildNode(smiNode)) {
	if (smiNode->nodekind & (SMI_NODEKIND_COLUMN | SMI_NODEKIND_SCALAR)
	    && (smiNode->access == SMI_ACCESS_READ_ONLY
		|| smiNode->access == SMI_ACCESS_READ_WRITE)) {

	    smiType = smiGetNodeType(smiNode);
	    if (!smiType) {
		continue;
	    }

	    if (isIndex(groupNode, smiNode)) {
		continue;
	    }
	    
	    cName = translate(smiNode->name);
	    fprintf(f,
		    "        if (vb->id_len > %u && vb->id[%u] == %u) {\n",
		    smiNode->oidlen, smiNode->oidlen - 1,
		    smiNode->oid[smiNode->oidlen - 1]);
	    switch (smiType->basetype) {
	    case SMI_BASETYPE_INTEGER32:
	    case SMI_BASETYPE_ENUM:
		fprintf(f,
			"            %s->%s = &(vb->syntax.i32[0]);\n",
			cGroupName, cName);
		break;
	    case SMI_BASETYPE_UNSIGNED32:
		fprintf(f,
			"            %s->%s = &(vb->syntax.ui32[0]);\n",
			cGroupName, cName);
		break;
	    case SMI_BASETYPE_INTEGER64:
		fprintf(f,
			"            %s->%s = &(vb->syntax.i64[0]);\n",
			cGroupName, cName);
		break;
	    case SMI_BASETYPE_UNSIGNED64:
		fprintf(f,
			"            %s->%s = &(vb->syntax.ui64[0]);\n",
			cGroupName, cName);
		break;
	    case SMI_BASETYPE_OCTETSTRING:
	    case SMI_BASETYPE_BITS:
		maxSize = getMaxSize(smiType);
		minSize = getMinSize(smiType);
		if (minSize != maxSize) {
		    fprintf(f,
			    "            %s->_%sLength = vb->syntax_len;\n",
			    cGroupName, cName);
		}
		fprintf(f,
			"            %s->%s = vb->syntax.uc;\n",
			cGroupName, cName);
		break;
	    case SMI_BASETYPE_OBJECTIDENTIFIER:
		fprintf(f,
			"            %s->_%sLength = vb->syntax_len / sizeof(guint32);\n",
			cGroupName, cName);
		fprintf(f,
			"            %s->%s = vb->syntax.ui32;\n",
			cGroupName, cName);
		break;
	    default:
		break;
	    }
	    fprintf(f,
		    "        }\n");
	    xfree(cName);
	}
    }

    xfree(cGroupName);
}



static void printAssignMethod(FILE *f, SmiModule *smiModule,
			      SmiNode *groupNode)
{
    char *cModuleName, *cGroupName;
    int i;

    cModuleName = translateLower(smiModule->name);
    cGroupName = translate(groupNode->name);

    if (groupNode->nodekind == SMI_NODEKIND_ROW) {
	printUnpackMethod(f, groupNode);
    }
    
    fprintf(f,
	    "static %s_t *\n"
	    "assign_%s(GSList *vbl)\n"
	    "{\n"
	    "    GSList *elem;\n"
	    "    %s_t *%s;\n"
	    "    char *p;\n",
	    cGroupName, cGroupName, cGroupName, cGroupName);

    fprintf(f, "    static guint32 const base[] = {");
    for (i = 0; i < groupNode->oidlen; i++) {
	fprintf(f, "%s%u", i ? ", " : "", groupNode->oid[i]);
    }
    fprintf(f, "};\n\n");

    fprintf(f,
	    "    %s = %s_new_%s();\n"
	    "    if (! %s) {\n"
	    "        return NULL;\n"
	    "    }\n"
	    "\n",
	    cGroupName, cModuleName, cGroupName, cGroupName);

    fprintf(f,
	    "    p = (char *) %s + sizeof(%s_t);\n"
	    "    * (GSList **) p = vbl;\n"
	    "\n", cGroupName, cGroupName);

    if (groupNode->nodekind == SMI_NODEKIND_ROW) {
	fprintf(f,
		"    if (unpack_%s((GSnmpVarBind *) vbl->data, %s) < 0) {\n"
		"        g_warning(\"illegal %s instance identifier\");\n"
		"        g_free(%s);\n"
		"        return NULL;\n"
		"    }\n\n",
		cGroupName, cGroupName, cGroupName, cGroupName);
    }
    
    fprintf(f,
	    "    for (elem = vbl; elem; elem = g_slist_next(elem)) {\n"
	    "        GSnmpVarBind *vb = (GSnmpVarBind *) elem->data;\n"
	    "        if (vb->type == G_SNMP_ENDOFMIBVIEW\n"
            "            || (vb->type == G_SNMP_NOSUCHOBJECT)\n"
            "            || (vb->type == G_SNMP_NOSUCHINSTANCE)) {\n"
            "            continue;\n"
	    "        }\n"
	    "        if (memcmp(vb->id, base, sizeof(base)) != 0) {\n"
	    "            continue;\n"
	    "        }\n");
    
    printVariableAssignement(f, groupNode);
    
    fprintf(f,
	    "    }\n"
	    "\n"
	    "    return %s;\n"
	    "}\n"
	    "\n", cGroupName);

    xfree(cGroupName);
    xfree(cModuleName);
}
 



static void printAddVarBind(FILE *f, SmiNode *smiNode, SmiNode *groupNode)
{
    SmiType *smiType;
    char *cName, *cGroupName;

    if (! (smiNode->nodekind & (SMI_NODEKIND_COLUMN | SMI_NODEKIND_SCALAR))
	&& (smiNode->access != SMI_ACCESS_READ_ONLY
	    && smiNode->access != SMI_ACCESS_READ_WRITE)) {
	return;
    }
    
    /*
     * Suppress all INDEX objects as if they were not-accessible.
     */

    if (isIndex(groupNode, smiNode)) {
	return;
    }

    cName = translate(smiNode->name);
    cGroupName = translate(groupNode->name);
    smiType = smiGetNodeType(smiNode);
    if (smiType && smiType->basetype == SMI_BASETYPE_UNSIGNED64) {
	fprintf(f, "    if (s->version > G_SNMP_V1) {\n    ");
    }
    fprintf(f,
	    "    base[%u] = %u; stls_vbl_add_null(&in, base, %u);\n",
	    smiNode->oidlen - 1, smiNode->oid[smiNode->oidlen - 1],
	    smiNode->oidlen);
    if (smiType && smiType->basetype == SMI_BASETYPE_UNSIGNED64) {
	fprintf(f, "    }\n");
    }
    xfree(cGroupName);
    xfree(cName);
}



static void printGetTableMethod(FILE *f, SmiModule *smiModule,
				SmiNode *entryNode)
{
    SmiNode *smiNode, *tableNode;
    char    *cModuleName, *cEntryName, *cTableName;
    int     i;

    tableNode = smiGetParentNode(entryNode);
    if (! tableNode) {
	return;
    }

    cModuleName = translateLower(smiModule->name);
    cEntryName = translate(entryNode->name);
    cTableName = translate(tableNode->name);

    fprintf(f,
	    "int\n"
	    "%s_get_%s(host_snmp *s, %s_t ***%s)\n"
	    "{\n"
	    "    GSList *in = NULL, *out = NULL;\n",
	    cModuleName, cTableName, cEntryName, cEntryName);

    fprintf(f,
	    "    GSList *row;\n"
	    "    int i;\n");

    fprintf(f, "    static guint32 base[] = {");
    for (i = 0; i < entryNode->oidlen; i++) {
	fprintf(f, "%u, ", entryNode->oid[i]);
    }
    fprintf(f, "0};\n");

    fprintf(f,
	    "\n"
	    "    *%s = NULL;\n"
	    "\n",
	    cEntryName);

    for (smiNode = smiGetFirstChildNode(entryNode);
	 smiNode;
	 smiNode = smiGetNextChildNode(smiNode)) {
	printAddVarBind(f, smiNode, entryNode);
    }

    fprintf(f,
	    "\n"
	    "    out = stls_snmp_gettable(s, in);\n"
	    "    /* stls_vbl_free(in); */\n"
	    "    if (! out) {\n"
	    "        return -2;\n"
	    "    }\n"
	    "\n");
    fprintf(f,
	    "    *%s = (%s_t **) g_malloc0((g_slist_length(out) + 1) * sizeof(%s_t *));\n"
	    "    if (! *%s) {\n"
	    "        return -4;\n"
	    "    }\n"
	    "\n", cEntryName, cEntryName, cEntryName, cEntryName);
    
    fprintf(f,
	    "    for (row = out, i = 0; row; row = g_slist_next(row), i++) {\n"
	    "        (*%s)[i] = assign_%s(row->data);\n"
	    "    }\n"
	    "\n", cEntryName, cEntryName);
    
    fprintf(f,
	    "    return 0;\n"
	    "}\n"
	    "\n");

    xfree(cTableName);
    xfree(cEntryName);
    xfree(cModuleName);
}
 



static void printGetMethod(FILE *f, SmiModule *smiModule,
			   SmiNode *groupNode)
{
    SmiNode *smiNode;
    char    *cModuleName, *cGroupName;
    int     i;

    cModuleName = translateLower(smiModule->name);
    cGroupName = translate(groupNode->name);

    fprintf(f,
	    "int\n"
	    "%s_get_%s(host_snmp *s, %s_t %s%s)\n"
	    "{\n"
	    "    GSList *in = NULL, *out = NULL;\n",
	    cModuleName, cGroupName, cGroupName,
	    (groupNode->nodekind == SMI_NODEKIND_ROW) ? "***" : "**",
	    cGroupName);

    fprintf(f, "    static guint32 base[] = {");
    for (i = 0; i < groupNode->oidlen; i++) {
	fprintf(f, "%u, ", groupNode->oid[i]);
    }
    fprintf(f, "0};\n");

    fprintf(f,
	    "\n"
	    "    *%s = NULL;\n"
	    "\n",
	    cGroupName);

    for (smiNode = smiGetFirstChildNode(groupNode);
	 smiNode;
	 smiNode = smiGetNextChildNode(smiNode)) {
	printAddVarBind(f, smiNode, groupNode);
    }

    fprintf(f,
	    "\n"
	    "    out = stls_snmp_getnext(s, in);\n"
	    "    stls_vbl_free(in);\n"
	    "    if (! out) {\n"
	    "        return -2;\n"
	    "    }\n"
	    "\n");
    
    fprintf(f,
	    "    *%s = assign_%s(out);\n"
	    "\n", cGroupName, cGroupName);

    fprintf(f,
	    "    return 0;\n"
	    "}\n"
	    "\n");

    xfree(cGroupName);
    xfree(cModuleName);
}



static void printNewMethod(FILE *f, SmiModule *smiModule,
			   SmiNode *groupNode)
{
    char *cModuleName, *cGroupName;

    cModuleName = translateLower(smiModule->name);
    cGroupName = translate(groupNode->name);

    fprintf(f,
	    "%s_t *\n"
	    "%s_new_%s()\n"
	    "{\n"
	    "    %s_t *%s;\n"
	    "\n",
	    cGroupName, cModuleName, cGroupName, cGroupName, cGroupName);

    fprintf(f,
	    "    %s = (%s_t *) g_malloc0(sizeof(%s_t) + sizeof(gpointer));\n"
	    "    return %s;\n"
	    "}\n"
	    "\n",
	    cGroupName, cGroupName, cGroupName, cGroupName);

    xfree(cGroupName);
    xfree(cModuleName);
}



static void printFreeTableMethod(FILE *f, SmiModule *smiModule,
				 SmiNode *groupNode)
{
    SmiNode *tableNode;
    char *cModuleName, *cGroupName, *cTableName;

    tableNode = smiGetParentNode(groupNode);
    if (! tableNode) {
	return;
    }

    cModuleName = translateLower(smiModule->name);
    cGroupName = translate(groupNode->name);
    cTableName = translate(tableNode->name);

    fprintf(f,
	    "void\n"
	    "%s_free_%s(%s_t **%s)\n"
	    "{\n"
	    "    int i;\n"
	    "\n",
	    cModuleName, cTableName, cGroupName, cGroupName);

    fprintf(f,	    
	    "    if (%s) {\n"
	    "        for (i = 0; %s[i]; i++) {\n"
	    "            %s_free_%s(%s[i]);\n"
	    "        }\n"
	    "        g_free(%s);\n"
	    "    }\n"
	    "}\n"
	    "\n",
	    cGroupName, cGroupName, cModuleName,
	    cGroupName, cGroupName, cGroupName);

    xfree(cTableName);
    xfree(cGroupName);
    xfree(cModuleName);
}
 



static void printFreeMethod(FILE *f, SmiModule *smiModule,
			    SmiNode *groupNode)
{
    char *cModuleName, *cGroupName;

    cModuleName = translateLower(smiModule->name);
    cGroupName = translate(groupNode->name);

    fprintf(f,
	    "void\n"
	    "%s_free_%s(%s_t *%s)\n"
	    "{\n"
	    "    GSList *vbl;\n"
	    "    char *p;\n"
	    "\n",
	    cModuleName, cGroupName, cGroupName, 
	    cGroupName);

    fprintf(f,	    
	    "    if (%s) {\n"
	    "        p = (char *) %s + sizeof(%s_t);\n"
	    "        vbl = * (GSList **) p;\n"
	    "        stls_vbl_free(vbl);\n"
	    "        g_free(%s);\n"
	    "    }\n"
	    "}\n"
	    "\n",
	    cGroupName, cGroupName, cGroupName, cGroupName);

    xfree(cGroupName);
    xfree(cModuleName);
}
 



static void printMethods(FILE *f, SmiModule *smiModule)
{
    SmiNode   *smiNode;
    int       cnt = 0;
    
    for (smiNode = smiGetFirstNode(smiModule, SMI_NODEKIND_ANY);
	 smiNode;
	 smiNode = smiGetNextNode(smiNode, SMI_NODEKIND_ANY)) {
	if (isGroup(smiNode) && isAccessible(smiNode)) {
	    cnt++;
	    printNewMethod(f, smiModule, smiNode);
	    printAssignMethod(f, smiModule, smiNode);
	    if (smiNode->nodekind == SMI_NODEKIND_ROW) {
		printGetTableMethod(f, smiModule, smiNode);
	    } else {
		printGetMethod(f, smiModule, smiNode);
	    }
	    printFreeMethod(f, smiModule, smiNode);
	    if (smiNode->nodekind == SMI_NODEKIND_ROW) {
		printFreeTableMethod(f, smiModule, smiNode);
	    }
	}
    }
    
    if (cnt) {
	fprintf(f, "\n");
    }
}



static void dumpStubs(SmiModule *smiModule, char *baseName)
{
    FILE *f;

    f = createFile(baseName, ".c");
    if (! f) {
        return;
    }

    printTopComment(f, smiModule);

    fprintf(f,
	    "#include \"%s.h\"\n"
	    "\n",
	    baseName);

    printEnumTables(f, smiModule);
    printMethods(f, smiModule);
    
    fclose(f);
}



static void dumpStools(int modc, SmiModule **modv, int flags, char *output)
{
    char	*baseName;
    int		i;

    if (flags & SMIDUMP_FLAG_UNITE) {
	/* not implemented yet */
    } else {
	for (i = 0; i < modc; i++) {
	    baseName = output ? output : translateFileName(modv[i]->name);
	    dumpHeader(modv[i], baseName);
	    dumpStubs(modv[i], baseName);
	    if (! output) xfree(baseName);
	}
    }

}



void initStools()
{
    static SmidumpDriver driver = {
	"stools",
	dumpStools,
	SMI_FLAG_NODESCR,
	SMIDUMP_DRIVER_CANT_UNITE,
	"ANSI C manager stubs for the stools package",
	NULL,
	NULL
    };

    smidumpRegisterDriver(&driver);
}
